(*
**
** Strong normalization for simply typed lambda calculus
**
*)

(*
** Interpretation of types 
*)

(* Int : Ty->(Tm->Set)

rec Int ty0 = SN
  | Int (arr s t) = ARR (Int s) (Int t)
*)

[Int = RecTyN SN 
	([_:Ty][Int_s:Tm->Set][_:Ty][Int_t:Tm->Set]ARR Int_s Int_t)];

Goal {t:Ty}CR (Int t);
Refine RecTy [t:Ty]CR (Int t);
Refine CR_SN;
intros;Refine CR_ARR;Immed;
Save CR_Int;

(* 
** Interpretation of judgements
*)

(* Mod : {n|Nat}(Con n)->Tm->Ty->Set

rec Mod zero empty M T = (Int T) M
  | Mod (succ n) (v_cons S G) M T = 
       	{N:Tm}(Int S N)->(Mod G (subst0 M (rep_weak0 N n) T)
	
*)

[Mod[n|Nat][G:Con n][M:Tm][T:Ty] =
	RecVecNN 
	  ([M:Tm] (Int T) M)
	  ([S:Ty][n|Nat][G:Con n][Mod_G:Tm->Prop]
		[M:Tm] {N:Tm}(Int S N)->(Mod_G (subst0 M (rep_weak0 N n))))
	  G M];

(*
** Soundness
*)

Goal {n|Nat}{G|Con n}{M|Tm}{T|Ty}(Der G M T)->(Mod G M T);
intros;
Refine RecDerN [n|Nat][G:Con n][M|Tm][T|Ty]Mod G M T;
(*** Var ***)
intros;
Refine RecFin [n|Nat][i:Fin n]{G:Con n}Mod G (var (Fin2Nat i)) (v_nth i G);
(* var zero *)
Refine RecVecSucc [n:Nat][G:Con (succ n)]Mod G (var zero) (v_hd G);
Intros S n2 G2 N S_N;
Refine RecVec [n2|Nat][G2 : Vec Ty n2]Mod G2 (rep_weak0 N n2) S;
Intros; Refine S_N;
Intros;Refine EQ_rewrite' ? ([M:Tm]Mod l M S);
Next +1;Refine subst_weak_lem zero; Refine H1;
(* var (succ i) *)
intros;
Refine RecVec1 [G2 : Con (succ n2)]
	Mod G2 (var (Fin2Nat (f_succ m))) (v_nth (f_succ m) G2);
Intros;Refine H1;
(*** App ***)
intros;
Refine RecVec [n|Nat][G1:Con n]
	{M1,N:Tm}(Mod G1 M1 (arr s t))->(Mod G1 N s)->Mod G1 (app M1 N) t;
(* G = empty *)
intros;Refine H5;Refine H6;
(* G = S::G' *)
Intros S m G' _______;Refine H5;Refine H6;Refine H8;Refine H7;Refine H8;
Immed;
(*** Lam ***)
intros;
Refine RecVec [n1|Nat][G1:Con n1]
	{M1:Tm}(Mod (v_cons s G1) M1 t)->(Mod G1 (lam M1) (arr s t));
(* G = empty *)
intros;Refine Lam_sound;Refine CR_Int;Refine CR_Int;Refine H3;
(* G = S::G' *)
Intros;Refine H3;Intros;
Equiv 
  Mod l (subst0 (subst (succ zero) M11 (rep_weak0 N n2)) (rep_weak0 N1 n2)) t;
Refine EQ_rewrite' ? ([M:Tm]Mod l M t);Next +1;
Refine subst_subst_lem zero; Refine H4; Immed;
(* END of CASES *)
Immed;Immed;
Save Int_sound;

(*
** Strong Normalisation
*)

Goal {n|Nat}{G|Con n}{M|Tm}{T|Ty}(Der G M T)->(SN M);
intros;
Claim (Mod G M T)->(SN M);Refine ?+1;Refine Int_sound;Immed;
Refine RecVec [n|Nat][G:Con n]{M:Tm}(Mod G M T)->(SN M);
(* G = empty *)
intros;Claim CR (Int T);Refine fst ?+1;Refine H1;Refine CR_Int;
(* G = S::G' *)
intros S m G' ___;
Refine SNvar m; Refine EQ_rewrite ? [X:Tm]SN (subst0 M1 X);
Next +1;Refine rep_weak0_lem;
Refine H1;Refine H2;Refine CR_nonEmpty;Refine CR_Int;
Save snorm;
